@@ -1,3 +1,21 @@
+/* 
+Copyright (C) 2003-2004 Simon N. Wood  simon@stats.gla.ac.uk
+
+This program is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License
+as published by the Free Software Foundation; either version 2
+of the License, or (at your option) any later version.
+
+This program is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+GNU General Public License for more details.
+(www.gnu.org/copyleft/gpl.html)
+
+You should have received a copy of the GNU General Public License
+along with this program; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
+USA. */
 #include <stdio.h>
 #include <stdlib.h>
 #include <math.h>
@@ -156,7 +174,7 @@ void fit_magic(double *X,double *sp,double **S,double *H,double *gamma,double *s
    to simply QR decompose [R',S^{1/2}']' at this point, attempt to estimate the rank of the problem 
    truncate the parameter space and proceed. However, QR is not a fool proof method for rank 
    estimation, so this routine actually uses a singular value decomposition, which is 
-   probably even proof against Tom Aitchison.  
+   probably even proof against Comrade Napoleon.  
 */
 
 { double *St,*p,*p1,xx,*R,*Vt,*a,trA,yAy,yAAy;
@@ -308,8 +326,8 @@ void magic_gH(double *U1U1,double **M,double **K,double *VS,double **My,double *
   }
 }
 
-void magic(double *y,double *X,double *sp,double *S,double *H,double *gamma,double *scale,int *control,
-           int *cS,double *rank_tol,double *tol,double *b,double *rV) 
+void magic(double *y,double *X,double *sp,double *def_sp,double *S,double *H,double *gamma,double *scale,
+           int *control,int *cS,double *rank_tol,double *tol,double *b,double *rV) 
 
 /* Maximally stable multiple gcv/ubre optimizer, based on pivoted QR decomposition and SVD, but without 
    a line search. At each point in the smoothing parameter space, the numerical rank of the problem 
@@ -323,6 +341,7 @@ void magic(double *y,double *X,double *sp,double *S,double *H,double *gamma,doub
    y - an n dimensional response vector
    X - an n by q model matrix
    sp - an m-array of smoothing parameters (any -ve => autoinitialize)
+   def_sp - an array of default values for sp's (any -ve => set up internally)
    b - a q dimensional parameter vector
    S - an array of dimension q columns of square roots of the m S_i penalty matrices. There are cS[i]
        columns for the ith penalty, and they are packed starting from i=0.
@@ -381,7 +400,7 @@ void magic(double *y,double *X,double *sp,double *S,double *H,double *gamma,doub
 
  */
 { int *pi,*pivot,q,n,autoinit,left,ScS,m,i,j,tp,k,use_sd=0,rank,converged,iter=0,ok,
-      gcv,try,fit_call=0,step_fail=0,max_half,*spok,*dir_sp,maxit;
+      gcv,try,fit_call=0,step_fail=0,max_half,*spok,/* *dir_sp,*/maxit,def_supplied;
   double *p,*p1,*p2,*tau,xx,*y1,*y0,yy,**Si=NULL,*work,score,*sd_step,*n_step,*U1,*V,*d,**M,**K,
          *VS,*U1U1,**My,**Ky,**yK,*dnorm,*ddelta,**d2norm,**d2delta,norm,delta,*grad,**hess,*nsp,
          min_score,*step,d_score=1e10,*ev=NULL,*u,msg=0.0,Xms,*rSms,*bag,*bsp,sign;
@@ -433,8 +452,11 @@ void magic(double *y,double *X,double *sp,double *S,double *H,double *gamma,doub
   }
    
   /* now get the initial smoothing parameter estimates \propto 1/tr(S_i) */
+
   autoinit=0;for (p=sp;p<sp+m;p++) if (*p <=0.0) { autoinit=1;break;} /* autoinitialize s.p.s? */ 
-  if (m>0)
+  def_supplied=1; for (p=def_sp;p<def_sp+m;p++) if (*p <=0.0) { def_supplied=0;break;} 
+
+  if (m>0&&!def_supplied) /* generate default sp's */
   { rSms=(double *)calloc((size_t)m,sizeof(double));
     /* first get some sort of norm for X */
     Xms=0.0;for (j=0;j<q;j++) for (i=0;i<=j;i++) { xx=X[i+n*j];Xms+=xx*xx;}
@@ -442,11 +464,14 @@ void magic(double *y,double *X,double *sp,double *S,double *H,double *gamma,doub
     for (i=0;i<m;i++)
     { for (xx=0.0,p=Si[i];p<Si[i]+q*q;p+=q+1) xx += *p;
       rSms[i]=xx/(q*cS[i]);
-      if (autoinit) sp[i]=log(Xms/rSms[i]); 
-      else sp[i]=log(sp[i]); /* transform smoothing parameters into log space! */
-    }
+      def_sp[i]=Xms/rSms[i]; 
+    }  
   } else { Xms=0.0;rSms=NULL;}
-  
+
+  if (autoinit) for (i=0;i<m;i++) sp[i]=log(def_sp[i]);
+  else for (i=0;i<m;i++) sp[i]=log(sp[i]);  
+
+/*  for (i=0;i<m;i++) Rprintf("%g  ",exp(sp[i]));Rprintf("\n");*/
 
   y1=(double *)calloc((size_t)q,sizeof(double)); /* Storage for U_1'Q_1'y */
   U1=(double *)calloc((size_t)(q*q),sizeof(double));
@@ -464,13 +489,13 @@ void magic(double *y,double *X,double *sp,double *S,double *H,double *gamma,doub
     u=(double *)calloc((size_t)(m*m),sizeof(double));
     U1U1=(double *)calloc((size_t)(q*q),sizeof(double));
     spok=(int *)calloc((size_t)m,sizeof(int));
-    dir_sp=(int *)calloc((size_t)m,sizeof(int));
+    /*dir_sp=(int *)calloc((size_t)m,sizeof(int));*/
     bsp=(double *)calloc((size_t)m,sizeof(double));
     bag=(double *)calloc((size_t)m,sizeof(double));
     } else 
     { M=K=My=Ky=yK=hess=d2norm=d2delta=NULL;
       VS=grad=dnorm=ddelta=nsp=ev=u=U1U1=bsp=bag=NULL;
-      spok=dir_sp=NULL;
+      spok=NULL;/*dir_sp=NULL;*/
     }
 
   fit_magic(X,sp,Si,H,gamma,scale,control,*rank_tol,yy,y0,y1,U1,V,d,b,&score,&norm,&delta,&rank);
@@ -484,7 +509,10 @@ void magic(double *y,double *X,double *sp,double *S,double *H,double *gamma,doub
     xx=1e-4*(1+fabs(score));
     ok=1;
     /* reset to default any sp w.r.t. which score is flat */
-    for (i=0;i<m;i++) if (fabs(grad[i])<xx) {sp[i]=log(Xms/(rSms[i]));ok=0;} 
+    for (i=0;i<m;i++) if (fabs(grad[i])<xx) 
+    { sp[i]=log(def_sp[i]);ok=0;
+    /*  Rprintf("Resetting sp[%d]\n",i);*/
+    } 
     if (!ok) 
     { fit_magic(X,sp,Si,H,gamma,scale,control,*rank_tol,yy,y0,y1,U1,V,d,b,&score,&norm,&delta,&rank);
       fit_call++;
@@ -496,7 +524,7 @@ void magic(double *y,double *X,double *sp,double *S,double *H,double *gamma,doub
   sd_step=(double *)calloc((size_t)m,sizeof(double));
   n_step=(double *)calloc((size_t)m,sizeof(double));
  
-  if (autoinit)
+  if (autoinit&&!def_supplied)
   { /* second guesses are scale*rank(S_i) / b'S_ib */
     for (p=S,k=0;k<m;k++)
     { for (j=0;j<cS[k];j++)
@@ -507,6 +535,7 @@ void magic(double *y,double *X,double *sp,double *S,double *H,double *gamma,doub
       sd_step[k]=log(*scale * cS[k]/xx)-sp[k]; 
     }
     use_sd=1;
+    /*  Rprintf("Using second guess\n");*/
   }
   
   /* Now do smoothing parameter estimation if there are any to estimate */
@@ -515,7 +544,9 @@ void magic(double *y,double *X,double *sp,double *S,double *H,double *gamma,doub
     while (!converged)
     { iter++;
       if (iter>200) error("magic, the gcv/ubre optimizer, failed to converge after 200 iterations.");
-      if (iter>1||autoinit) ok=1; else ok=0;try=0;
+      if (iter>1||(autoinit&&!def_supplied)) ok=1; /* try out step */
+      else ok=0; /* no step to try yet */
+      try=0; 
       if (use_sd) step=sd_step; else step=n_step;
       while (ok) /* try out step, shrinking it if need be */
       { try++; if (try==4&&!use_sd) {use_sd=1;step=sd_step;}
@@ -527,9 +558,11 @@ void magic(double *y,double *X,double *sp,double *S,double *H,double *gamma,doub
           d_score=min_score-score;
           min_score=score;
           for (i=0;i<m;i++) sp[i]=nsp[i];
+     /*   for (i=0;i<m;i++) Rprintf("%g  ",exp(sp[i]));Rprintf("\n");*/
         } else
         for (i=0;i<m;i++) step[i]/=2;
-        if (try==(max_half-1)&&ok) for (i=0;i<m;i++) step[i]=0.0; /* reset sp's to best so far before giving up */
+        if (try==(max_half-1)&&ok) 
+        for (i=0;i<m;i++) step[i]=0.0; /* reset sp's to best so far before giving up */
         if (try==max_half) {ok=0;} /* give up */
       }
       if (iter>3) /* test for convergence */
@@ -576,9 +609,10 @@ void magic(double *y,double *X,double *sp,double *S,double *H,double *gamma,doub
     }
     fit_magic(X,sp,Si,H,gamma,scale,control,*rank_tol,yy,y0,y1,U1,V,d,b,&score,&norm,&delta,&rank);
     /*Rprintf("\n Rank at final call = %d",rank);*/
+   
     /* free search related memory */
     free2d(Si);free2d(M);free2d(K);free2d(My);free2d(Ky);free2d(yK);free2d(hess);
-    free2d(d2norm);free2d(d2delta);free(U1U1);free(rSms);
+    free2d(d2norm);free2d(d2delta);free(U1U1);free(rSms);free(u);
     free(VS);free(grad);free(dnorm);free(ddelta);free(nsp);free(ev);
     free(bsp);free(bag);free(spok);
   }   
@@ -596,7 +630,6 @@ void magic(double *y,double *X,double *sp,double *S,double *H,double *gamma,doub
   *tol = msg; /* the root mean square gradient at convergence */  
   control[0]=rank; /* problem rank at convergence */
   control[1]=1-step_fail; /* 1 for true convergence, 0  for step failure */
-  use_sd=0;for (p=ev;p<ev+m;p++) if (*p<0.0) {use_sd=1;break;} /* check hessian +ve def */
   control[2]=1-use_sd; /* 1 if Hessian +ve definite, 0 otherwise */
   control[3]=iter; /* iterations used */
   control[4]=fit_call; /* number of evaluations of GCV/UBRE score */
